/*
* Dateiname.... lowLevel.c
* erstellt am.. 2016-10-30
* gendert am.. 2017-02-24
* Autor........ Clemens B Fahr
* email........ clemens.fahr[at]fahrnet.de
* Kommentar.... based on RFM_wireless_4CH_sensor_02_mod
* ............. modify LED blink interval times on/off
* ............. modify moment to power on/off the radio
* ............. modify the RFM init routine; shift parts to the RFM transmitt
* ............. wakeup issue bug-fixed
*/

// includes
//------------------------------------------------------------------
#include "lowlevel.h"
#include "global.h"

void LOWLEVEL_init_all(void)
{	
	// write oscillator calibration byte
	//OSCCAL = 0x4C; 	// has to be determined by AVR Studio
	DDRA = 0;
	DDRB = 0;
	
	// clock prescaler: activate in case of reducing the speed of a
    // external clock source; only functions if CLKDIV8-fuse is disabled
	#ifdef NOCRYSTAL
	CLKPR = (0x01 << CLKPCE);	// enable to write the prescaler
	CLKPR = CLK_PRESCALER;	// write the prescaler value
	#endif
	
	// start basic initialisation of data direction, port setting and pull-up
	// ADC ports
	SENS_EN(HIGH);
	SENS_EN_DDR;
	
	// others
	#ifndef DEBUGON
		LED1_DDR;
		LED1(OFF); // status LED off
	#endif
	#ifndef DEBUGADCON
		LED1_DDR;
		LED1(OFF); // status LED off
	#endif
	#ifndef DEBUGADCTHRESHOLD
		LED1_DDR;
		LED1(OFF); // status LED off
	#endif
	#ifdef DEBUGON
		DEBUG1_DDR;
		DEBUG1(LOW);
	#endif
	#ifdef DEBUGADCON
		DEBUG1_DDR;
		DEBUG1(LOW);
	#endif
	#ifdef DEBUGADCTHRESHOLD
		DEBUG1_DDR;
		DEBUG1(LOW);
	#endif
}

void LOWLEVEL_init_ADC_buffer(uint8_t numba_bytes)
{
        uint8_t i = 0;
        for (i=0;i<numba_bytes;i++) {
                ADC_previous_value[i] = 0xff;
                ADC_TX_holdoff[i] = FALSE;
        }
}

void LOWLEVEL_init_RFM(void)
{
	// init pins and their data direction, init pin level
	RFM_SEL(HIGH);
	RFM_SDI(HIGH);
	RFM_SCK(LOW);
	RFM_FSK(LOW);
	RFM_SEL_DDR;
	RFM_SDI_DDR;
	RFM_IRQ_DDR;
	RFM_SCK_DDR;
	RFM_FSK_DDR;
	
	//send_command_RFM(RFM_CMD_READSTATUS); // seems to cause some disruption - 20170224
}

void LOWLEVEL_sleep_RFM(uint16_t sleep_time)
{	
	send_command_RFM(RFM_CMD_PWRMNG); // clear all - power off
	
	if ( sleep_time == REPEAT_SLEEP ) {
		// modify the sleep time for the repeat-telegram timeout with a little jitter,
		// based on the unique device ID - the basic sleep time will be increased by multiples of 512ms
		// with modulus 16 of the lower unique device ID counter
		//sleep_time += (eeprom_read_byte(&DEVICE_UID2) % 16); // NEEDS to be checked!!!
		;
	}
	
	// programme the wake-up timer of the RFM02
	send_command_RFM(sleep_time);
	send_command_RFM(RFM_CMD_LBDSET);
	send_command_RFM(RFM_CMD_PWRMNG | (1 << ENABLE_WAKEUP_TIMER_ET) ); // enable timer wakeup 
	send_command_RFM(RFM_CMD_SLEEP);
}

uint16_t LOWLEVEL_status_register_read(void)
{
	return send_command_RFM(RFM_CMD_READSTATUS);
}

uint16_t send_command_RFM(uint16_t aCommand)
{
	uint8_t i;
	uint16_t temp=0;
	
	RFM_SCK(LOW);
	nop();
	RFM_SEL(LOW);
	nop();
	for (i=0;i<16;i++) {
		temp <<= 1;
		if (RFM_IRQ_HIGH) {
			temp |= 0x0001;
		}
		if (aCommand & 0x8000) {
			RFM_SDI(HIGH);
		}
		else {
			RFM_SDI(LOW);
		}
		nop();
		RFM_SCK(HIGH);
		nop();
		RFM_SCK(LOW);
		nop();
		aCommand <<= 1;
	}
	RFM_SEL(HIGH);
	nop();
	return temp;
}

void LOWLEVEL_send_telegram_RFM(uint8_t NumbaDataBytes)
{
	
	send_command_RFM(RFM_CMD_CONFIGSET);
	send_command_RFM(RFM_CMD_FREQSET);
	send_command_RFM(RFM_CMD_DATARATE);
	send_command_RFM(RFM_CMD_HALFDATARATE);
	send_command_RFM(RFM_CMD_PWRSET);
	send_command_RFM(RFM_CMD_PLLSET);
	
	send_command_RFM(RFM_CMD_LBDSET | (1 << ENABLE_TX_BIT_SYNCHRO_EBS));
	send_command_RFM(RFM_CMD_PWRMNG);		// clear all
	send_command_RFM(RFM_CMD_PWRMNG | (1 << ENABLE_CRYSTAL_EX) | (1 << ENABLE_SYNTHESIZER_ES) | (1 << ENABLE_AMPLIFIER_EA) ); // turn on crystal, synthesizer and amplifier 
	//LOWLEVEL_delay_1ms(1); // wait for PLL, CLK etc. to lock - not needed, as nIRQ will indicate the moment, the module is ready for transmission

	uint8_t i, ii, TX_byte;
	for (i=0;i<NumbaDataBytes;i++) {
		TX_byte = TX_buffer[i];
		for (ii=0;ii<8;ii++) {
			while (RFM_IRQ_HIGH);
			while (!(RFM_IRQ_HIGH));
			if (TX_byte&0x80) {
				RFM_FSK(HIGH);
			}
			else {
				RFM_FSK(LOW);
			}
			TX_byte <<= 1;
		}
	}
}

void LOWLEVEL_debugOut(uint8_t aByte)
{
	uint8_t i;
	
	// send the start bit
	DEBUG1(HIGH);	
	UART_bit_delay();
	// send the aByte payload
	for (i=0;i<8;i++) {
		if (aByte & 0x01) {
			DEBUG1(LOW);
		}
		else {
			DEBUG1(HIGH);
		}
		UART_bit_delay();
		aByte >>= 1;
	}
	// send the stop bit
	DEBUG1(LOW);
	UART_bit_delay();
	UART_bit_delay();
	UART_bit_delay();
}

void UART_bit_delay(void)
{
	run_timer0(PRELOAD_BIT_T0,CLK_SRC_BIT_T0,OFF);
	do
	{	//while timer does not overflow stay here = delay
	}
	while ((TIFR & (0x01 << TOV0)) == 0);
	stop_timer0();
}

// some default timer routines
void LOWLEVEL_delay_1ms(uint16_t multiplier)
{
	// produces a (1 * multiplier)ms delay
	uint16_t i;
	
	for (i=0;i<multiplier;i++) {
		run_timer1(PRELOAD_1MS_T1,CLK_SRC_1MS_T1,OFF);
		do
		{	//while timer does not overflow stay here = delay
		}
		while ((TIFR & (0x01 << TOV1)) == 0);
		stop_timer1();
	}
}

void LOWLEVEL_delay_sleep_100ms(volatile uint8_t multiplier)
{
    // produces a (1 * multiplier) 100ms delay and sleeps for 100ms each
    volatile uint16_t i;
	
	for (i=0;i<multiplier;i++) {
        // enable to wake up on timer interrupt TOV1 event, need low level interrupt for timer interrupt TOV1
		// timer interrupt is supported when IDLE is issued as sleep mode
		GIMSK |= (0x01 << TOV1);
        run_timer1(PRELOAD_100MS_T1,CLK_SRC_100MS_T1,ON);
		sei();
		// enable the sleep mode function, set the sleep mode to POWER DOWN
		set_sleep_mode(SLEEP_MODE_IDLE);
		sleep_mode();
		// entrypoint after repeat-sleep wakeup is here
        cli();
        stop_timer1();
    }
}

void run_timer0(uint8_t preloadvalue, uint8_t clocksource, uint8_t IRQ_enabled)
{
	// used for the software UART only
	// clear interrupt flag TOV0
	TIFR = (0x01 << TOV0); // annotation: remember to NOT use the bit-"OR"!! Interrupt flags are cleared this way ONLY!!
	// preload timer counter and the compare match register for 8bit counter mode
	TCNT0L = preloadvalue;
	OCR0A = 0xFF;
	//TCNT0H = 0xFF;
	// setup timer/counter 1 control register
	TCCR0B |= clocksource;
	if ( IRQ_enabled ) {
		TIMSK |= (0x01 << TOIE0);
	}
	else {
		TIMSK &= ~(0x01 << TOIE0);
	}
}

void stop_timer0(void)
{
	// setup timer/counter 1 control register, stop clocking of timer
	// makes the timer stop
	TCCR0B &= TIMER0_CLK_OFF;
	// disables timer interrup
	TIMSK &= ~(0x01 << TOIE0);
	// preload timer counter to bottom
	TCNT0L = 0x00;
	OCR0A = 0x00;
	//TCNT0H = 0x00;
}

void run_timer1(uint8_t preloadvalue, uint8_t clocksource, uint8_t IRQ_enabled)
{
	// clear interrupt flag TOV1
	TIFR = (0x01 << TOV1); // annotation: remember to NOT use the bit-"OR"!! Interrupt flags are cleared this way ONLY!!
	// preload timer counter and the compare match register for 8bit counter mode
	OCR1C = 0xFF; // not sure, if this is truely needed
	TCNT1 = preloadvalue;
	// setup timer/counter 1 control register
	TCCR1B |= clocksource;
	if ( IRQ_enabled ) {
		TIMSK |= (0x01 << TOIE1);
	}
	else {
		TIMSK &= ~(0x01 << TOIE1);
	}
}

void stop_timer1(void)
{
	// setup timer/counter 1 control register, stop clocking of timer
	// makes the timer stop
	TCCR1B &= TIMER1_CLK_OFF;
	// disables timer interrup
	TIMSK &= ~(0x01 << TOIE1);
	// preload timer counter to bottom
	TCNT1 = 0x00;
}

void LOWLEVEL_blink_status_LED(volatile uint8_t numba_blink)
{
    volatile uint8_t i;

    for (i=0;i<numba_blink;i++) {
            LED1(ON);
            //LOWLEVEL_delay_1ms(300); // this is not energy sensitive
            LOWLEVEL_delay_sleep_100ms(1);
            LED1(OFF);
            //LOWLEVEL_delay_1ms(300); // this is not energy sensitive
            LOWLEVEL_delay_sleep_100ms(2);
	}
}

uint16_t LOWLEVEL_adcConvert(uint8_t ADC_channel, uint8_t reference_voltage)
{
	uint8_t i;
	uint16_t average = 0;	
	
	// setting up AD converter
	if (reference_voltage == 1) {
		// AD multiplexer, AREF as reference and single ended ADC (MUX5..0 = 00000)
		ADMUX = (0x01 << REFS0);
		ADCSRB = 0;
	}
	else if (reference_voltage == 0) {
		// AD multiplexer, use internal 2.56V voltage reference with AREF disconnected(!) and single ended ADC (MUX5..0 = 00000)
		ADMUX = (0x01 << REFS1);
		ADCSRB = (0x01 << REFS2);
	}
	// ADC control and status register
	// set ADC clock to CK/32 (roughly 63kHz @ 2MHz)
	ADCSRA = (0x01 << ADPS2) | (0x01 << ADPS0);

	// select the mux-channel according to ADC_channel; channel must be given acc to channel number ADC0..ADC11 (!) - don't mix it with port number
	ADMUX |= ADC_channel;
	
	// enable ADC and run
	ADCSRA |= (0x01 << ADEN);
	for (i=0;i<6;i++) {
		// clear interrupt flag by writing a one
		ADCSRA |= (0x01 << ADIF);
		// start ADC conversion
		ADCSRA |= (0x01 << ADSC);
		do
		{
		}
		while ((ADCSRA & (0x01 << ADIF)) == 0); //until the ADC interrupt flag is set
	}
	
	average = ADCL;
	average |= ADCH << 8;
	
	// disable ADC
	ADCSRA = 0;
	ADCSRB = 0;
	// clear the mux-channel selection
	ADMUX = 0;
	// clear the ADC interrupt flag
	ADCSRA = (0x01 << ADIF);
	
	// return value to function
	return average;
}